<?php

namespace Drupal\reviewer\Drush\Commands;

use Consolidation\OutputFormatters\StructuredData\RowsOfFields;
use Drupal\reviewer\Plugin\reviewer\Review\ReviewPluginInterface;
use Drupal\reviewer\Plugin\ReviewManagerInterface;
use Drupal\reviewer\Reviewer\IgnorerInterface;
use Drupal\reviewer\Reviewer\RunnerInterface;
use Drupal\reviewer\Reviewer\Status\StatusEvaluatorInterface;
use Drupal\reviewer\Reviewer\Status\StatusFactoryInterface;
use Drush\Attributes as CLI;
use Drush\Commands\DrushCommands;
use Symfony\Component\Console\Output\ConsoleOutput;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Drush commands for reviewer.
 */
final class ReviewerCommands extends DrushCommands {

  // phpcs:ignore Drupal.Commenting.FunctionComment.Missing
  public function __construct(
    private readonly IgnorerInterface $ignorer,
    private readonly ReviewManagerInterface $reviewManager,
    private readonly RunnerInterface $reviewRunner,
    private readonly StatusEvaluatorInterface $statusEvaluator,
    private readonly StatusFactoryInterface $statusFactory,
  ) {
    parent::__construct();
  }

  // phpcs:ignore Drupal.Commenting.FunctionComment.Missing
  public static function create(
    ContainerInterface $container,
  ): ReviewerCommands {
    return new ReviewerCommands(
      $container->get('reviewer.ignorer'),
      $container->get('plugin.manager.reviewer.review'),
      $container->get('reviewer.runner'),
      $container->get('reviewer.evaluator.status'),
      $container->get('reviewer.factory.status'),
    );
  }

  /**
   * Run reviews.
   *
   * @param string[] $ids
   * @param array<string, bool|string> $options
   *
   * @throws \Drupal\reviewer\Exception\CannotOverridePropertyException
   * @throws \Drupal\reviewer\Exception\NoChecklistsException
   * @throws \Drupal\reviewer\Exception\NotRunException
   * @throws \Drupal\reviewer\Exception\NotIgnoredException
   */
  #[CLI\Command(name: 'reviewer:run', aliases: ['rvr'])]
  #[CLI\Argument(name: 'ids', description: 'Review IDs to run, separated by spaces.')]
  #[CLI\Option(name: 'review', description: 'Interactively review, prompting to ignore failures and errors.')]
  #[CLI\Option(name: 'review-new', description: 'Interactively review, prompting to ignore new failures and errors only.')]
  #[CLI\Option(name: 'review-reset', description: 'Interactively review, prompting to ignore all failures and errors. Any previously ignored items will no longer be ignored.')]
  #[ClI\Option(name: 'show-passed', description: 'Display passed tests in addition to failures and errors.')]
  #[ClI\Option(name: 'show-ignored', description: 'Display ignored tests in addition to failures and errors.')]
  #[ClI\Option(name: 'show-all', description: 'Display ignored and passed tests in addition to failures and errors.')]
  #[CLI\Option(name: 'window-width', description: 'Set the window width to constrain the output table size.')]
  #[CLI\Help(description: 'Run reviews.')]
  #[CLI\Usage(name: 'reviewer:run', description: 'Run all reviews.')]
  #[CLI\Usage(name: 'reviewer:run node', description: 'Run reviews for all node types.')]
  #[CLI\Usage(name: 'reviewer:run node:article,page paragraph:gallery', description: 'Run reviews for the article and page node types and the gallery paragraph type.')]
  public function run(
    array $ids,
    array $options = [
      'review' => FALSE,
      'review-new' => FALSE,
      'review-reset' => FALSE,
      'show-ignored' => FALSE,
      'show-passed' => FALSE,
      'show-all' => FALSE,
      'window-width' => '80',
    ],
  ): void {
    $do_review = $options['review'] || $options['review-new'] || $options['review-reset'];
    $do_review_new = (bool) $options['review-new'];
    $do_review_reset = (bool) $options['review-reset'];
    $show_ignored = $options['show-ignored'] || $options['show-all'];
    $show_passed = $options['show-passed'] || $options['show-all'];
    $show_all = (bool) $options['show-all'];

    $output = new ConsoleOutput();

    if (!\is_numeric($options['window-width'])) {
      $this->io()->warning(dt('Ignoring --window-width: Option value must be an integer.'));
      $window_width = 80;
    }
    else {
      $window_width = (int) $options['window-width'];
    }

    $reviews = [];
    if ($review_ids = $this->processReviewIds($ids)) {
      foreach ($review_ids as $review_id => $bundles) {
        if (!$this->reviewManager->hasDefinition($review_id)) {
          $this->io()->warning(dt('Review @id does not exist.', [
            '@id' => $review_id,
          ]));
        }

        $reviews = array_merge($reviews, $this->reviewRunner->runIds(
          [$review_id],
          $bundles,
        ));
      }
    }
    else {
      $reviews = $this->reviewRunner->runIds();
    }

    foreach ($reviews as $review) {
      $rows = [];

      if ($do_review_reset) {
        $this->ignorer->unignore($review->results()->individualResults());
      }

      if ($do_review) {
        foreach ($review->results()->individualResults() as $result) {
          if ($this->statusEvaluator->isPass($result->getStatus())) {
            continue;
          }

          $ignore = FALSE;
          $update = FALSE;
          $reason = '';

          if (
            !$do_review_new
            && $this->ignorer->isConfigIgnored($result)
          ) {
            $reason = $this->ignorer->ignoredReason($result);

            $ignore = $this->io()->confirm(dt('Keep ignoring result @id?' . PHP_EOL . ' Current ignored reason: @reason', [
              '@id' => $result->getId(),
              '@reason' => $reason,
            ]));

            $update = TRUE;
          }

          if (!$this->ignorer->isConfigIgnored($result)) {
            $ignore = $this->io()->confirm(
              dt('Ignore result @id?', ['@id' => $result->getId()]),
              FALSE,
            );

            $update = TRUE;
          }

          if (!$update) {
            continue;
          }

          if ($ignore) {
            $reason = (string) $this->io()->ask(
              dt('Provide a reason for ignoring @id.', [
                '@id' => $result->getId(),
              ]),
              $reason,
            );

            $this->ignorer->ignore($result, $reason);
            $result->overrideStatus($this->statusFactory->createIgnored($result->getStatus()));
            $result->setOverrideMessage($reason);
          }
          else {
            $this->ignorer->unignore($result);
            $result->overrideStatus($this->statusFactory->createUnignored($result->getStatus()));
            $result->setOverrideMessage('');
          }
        }
      }

      $results = match ($show_all) {
        TRUE => $review->results()->individualResults(),
        FALSE => $this->statusEvaluator->isPass($review->result()->getStatus())
          ? [$review->result()]
          : $review->results()->individualResults(),
      };

      foreach ($results as $result) {
        if ($show_passed && $this->statusEvaluator->isPass($result->getStatus())) {
          $rows[] = $result;
          continue;
        }

        if ($show_ignored && $this->ignorer->isIgnored($result)) {
          $rows[] = $result;
          continue;
        }

        if (
          !$this->statusEvaluator->isPass($result->getStatus())
          && !$this->ignorer->isIgnored($result)
        ) {
          $rows[] = $result;
        }
      }

      if ($rows) {
        $table = $this->io()->createTable()
          ->setVertical()
          ->setColumnWidth(0, max(0, $window_width - 4))
          ->setColumnMaxWidth(0, max(0, $window_width - 4))
          ->setHeaderTitle($review->getLabel())
          ->setHeaders([dt('Status'), dt('ID'), dt('Message')]);

        foreach ($rows as $row) {
          $table->addRow([
            $row->getStatus()->labelConsole(),
            $row->getId(),
            $row->message(),
          ]);
        }

        $output->writeln('');
        $table->render();
        $output->writeln('');
      }
    }
  }

  /**
   * List all available reviews.
   */
  #[CLI\Command(name: 'reviewer:list', aliases: ['rvl'])]
  #[CLI\Help(description: 'List available reviews.')]
  #[CLI\Usage(name: 'reviewer:list', description: 'List available reviews.')]
  #[CLI\DefaultTableFields(fields: ['id', 'label'])]
  #[CLI\FieldLabels(labels: ['id' => 'ID', 'label' => 'Label'])]
  public function list(): RowsOfFields {
    $this->io()->writeln('');
    $this->io()->writeln(dt('Available Reviews:'));
    return new RowsOfFields(array_map(
      fn(ReviewPluginInterface $review) => ['id' => $review->getPluginId(), 'label' => $review->getLabel()],
      $this->reviewManager->createAllInstances(),
    ));
  }

  /**
   * Process review IDs and return an array of review IDs and their bundles.
   *
   * @param string[] $review_ids
   *
   * @return array<string, string[]>
   *   Array keys are the review IDs with the values the bundles for each review
   *   ID, with an empty array for all bundles or a review which does not
   *   support bundles.
   */
  private function processReviewIds(array $review_ids): array {
    $processed_ids = [];
    foreach ($review_ids as $review_id) {
      $split_ids = explode(':', $review_id);
      $bundles = match (count($split_ids)) {
        1 => [],
        default => explode(',', $split_ids[1]),
      };
      $processed_ids = array_merge(
        $processed_ids,
        array_fill_keys(explode(',', $split_ids[0]), $bundles),
      );
    }
    return $processed_ids;
  }

  /**
   * {@inheritdoc}
   */
  #[\Override] protected function io(): SymfonyStyle {
    // @phpstan-ignore-next-line
    if (!$this->io instanceof SymfonyStyle) {
      $this->io = new SymfonyStyle($this->input(), $this->output());
    }
    return $this->io;
  }

}
